@@ -39,9 +39,6 @@ static GObjectClass *parent_class = NULL;
 static gint
 rstto_navigator_entry_name_compare_func(RsttoNavigatorEntry *a, RsttoNavigatorEntry *b);
 
-static void
-rstto_navigator_entry_free_pixbuf (RsttoNavigatorEntry *entry);
-
 enum
 {
     RSTTO_NAVIGATOR_SIGNAL_ENTRY_MODIFIED = 0,
@@ -55,7 +52,6 @@ struct _RsttoNavigatorEntry
 {
     ThunarVfsInfo       *info;
     GdkPixbuf           *thumb;
-    GdkPixbuf           *pixbuf;
     gdouble              scale;
     gboolean             fit_to_screen;
     ExifData            *exif_data;
@@ -101,6 +97,8 @@ rstto_navigator_init(RsttoNavigator *navigator)
     navigator->old_position = -1;
     navigator->timeout = 5000;
     navigator->album = FALSE;
+
+    navigator->factory = thunar_vfs_thumb_factory_new(THUNAR_VFS_THUMB_SIZE_NORMAL);
 }
 
 static void
@@ -197,7 +195,6 @@ rstto_navigator_jump_first (RsttoNavigator *navigator)
     if(navigator->file_iter)
     {
         navigator->old_position = rstto_navigator_get_position(navigator);
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
     }
     navigator->file_iter = g_list_first(navigator->file_list);
     if(navigator->file_iter)
@@ -212,7 +209,6 @@ rstto_navigator_jump_forward (RsttoNavigator *navigator)
     if(navigator->file_iter)
     {
         navigator->old_position = rstto_navigator_get_position(navigator);
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
         navigator->file_iter = g_list_next(navigator->file_iter);
     }
     if(!navigator->file_iter)
@@ -245,7 +241,6 @@ rstto_navigator_jump_back (RsttoNavigator *navigator)
     if(navigator->file_iter)
     {
         navigator->old_position = rstto_navigator_get_position(navigator);
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
         navigator->file_iter = g_list_previous(navigator->file_iter);
     }
     if(!navigator->file_iter)
@@ -268,7 +263,6 @@ rstto_navigator_jump_last (RsttoNavigator *navigator)
     if(navigator->file_iter)
     {
         navigator->old_position = rstto_navigator_get_position(navigator);
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
     }
     navigator->file_iter = g_list_last(navigator->file_list);
 
@@ -338,11 +332,6 @@ rstto_navigator_get_nth_file (RsttoNavigator *navigator, gint n)
 gint
 rstto_navigator_add (RsttoNavigator *navigator, RsttoNavigatorEntry *entry)
 {
-    if(navigator->file_iter)
-    {
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
-    }
-
     navigator->file_list = g_list_insert_sorted(navigator->file_list, entry, navigator->compare_func);
     if (!navigator->file_iter)
     {
@@ -366,7 +355,6 @@ rstto_navigator_remove (RsttoNavigator *navigator, RsttoNavigatorEntry *entry)
         if(navigator->file_iter->data == entry)
         {
             navigator->old_position = rstto_navigator_get_position(navigator);
-            rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
             navigator->file_iter = g_list_next(navigator->file_iter);
             if(!navigator->file_iter)
                 navigator->file_iter = g_list_first(navigator->file_list);
@@ -429,7 +417,6 @@ rstto_navigator_set_file (RsttoNavigator *navigator, gint n)
     if(navigator->file_iter)
     {
         navigator->old_position = rstto_navigator_get_position(navigator);
-        rstto_navigator_entry_free_pixbuf(navigator->file_iter->data);
     }
     navigator->file_iter = g_list_nth(navigator->file_list, n);
     if(navigator->file_iter)
@@ -586,87 +573,57 @@ rstto_navigator_entry_get_flip (RsttoNavigatorEntry *entry, gboolean horizontal)
 void
 rstto_navigator_entry_free(RsttoNavigatorEntry *nav_entry)
 {
-    if(nav_entry->pixbuf)
-    {
-        g_object_unref(nav_entry->pixbuf);
-    }
     if(nav_entry->thumb)
     {
-        g_object_unref(nav_entry->thumb);
+        gdk_pixbuf_unref(nav_entry->thumb);
+        nav_entry->thumb = NULL;
     }
     if(nav_entry->exif_data)
     {   
         exif_data_free(nav_entry->exif_data);
+        nav_entry->exif_data = NULL;
     }
     thunar_vfs_info_unref(nav_entry->info);
     g_free(nav_entry);
 }
 
 GdkPixbuf *
-rstto_navigator_entry_get_thumb(RsttoNavigatorEntry *entry, gint size)
+rstto_navigator_get_entry_thumb(RsttoNavigator *navigator, RsttoNavigatorEntry *entry, gint size)
 {
     if(entry->thumb)    
     {
         if(!(gdk_pixbuf_get_width(entry->thumb) == size || gdk_pixbuf_get_height(entry->thumb) == size))
         {
-            g_object_unref(entry->thumb);
-            gchar *filename = thunar_vfs_path_dup_string(entry->info->path);
-            entry->thumb = gdk_pixbuf_new_from_file_at_size(filename, size, size, NULL);
-            g_free(filename);
         }
     }
     else
     {
-        gchar *filename = thunar_vfs_path_dup_string(entry->info->path);
-        entry->thumb = gdk_pixbuf_new_from_file_at_size(filename, size, size, NULL);
-        g_free(filename);
-    }
-    return entry->thumb;
-}
-
-GdkPixbuf *
-rstto_navigator_entry_get_pixbuf(RsttoNavigatorEntry *entry)
-{
-    g_return_val_if_fail(entry, NULL);
-
-    if(!entry->pixbuf)
-    {
-        gchar *filename = thunar_vfs_path_dup_string(entry->info->path);
-        GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
-        if (pixbuf)
+        ThunarVfsInfo *info = rstto_navigator_entry_get_info(entry);
+        gchar *thumbnail = thunar_vfs_thumb_factory_lookup_thumbnail(navigator->factory, info);
+        if (thumbnail == NULL)
         {
-            entry->pixbuf = gdk_pixbuf_rotate_simple(pixbuf, entry->rotation);
-            gdk_pixbuf_unref(pixbuf);
-            if (entry->v_flipped)
+            GdkPixbuf *pixbuf = thunar_vfs_thumb_factory_generate_thumbnail(navigator->factory, info);
+            if (pixbuf != NULL)
             {
-                pixbuf = entry->pixbuf;
-                entry->pixbuf = gdk_pixbuf_flip(pixbuf, FALSE);
-                gdk_pixbuf_unref(pixbuf);
+                if (!thunar_vfs_thumb_factory_store_thumbnail(navigator->factory, pixbuf, info, NULL))
+                {
+                }
+                entry->thumb = gdk_pixbuf_scale_simple(pixbuf, size, size, GDK_INTERP_BILINEAR);
             }
-            if (entry->v_flipped)
+            else
             {
-                pixbuf = entry->pixbuf;
-                entry->pixbuf = gdk_pixbuf_flip(pixbuf, TRUE);
-                gdk_pixbuf_unref(pixbuf);
+                thumbnail = thunar_vfs_path_dup_string(info->path);
+                entry->thumb = gdk_pixbuf_new_from_file_at_size(thumbnail, size, size, NULL);
+                g_free(thumbnail);
             }
         }
         else
         {
-            entry->pixbuf = NULL;
+            entry->thumb = gdk_pixbuf_new_from_file_at_size(thumbnail, size, size, NULL);
+            g_free(thumbnail);
         }
-        g_free(filename);
-    }
-    return entry->pixbuf;
-}
-
-static void
-rstto_navigator_entry_free_pixbuf (RsttoNavigatorEntry *entry)
-{
-    if(entry->pixbuf)
-    {
-        g_object_unref(entry->pixbuf);
-        entry->pixbuf = NULL;
     }
+    return entry->thumb;
 }
 
 gint
