@@ -26,6 +26,10 @@
 
 struct _RsttoPictureViewerPriv
 {
+    GdkPixbufLoader  *loader;
+    GdkPixbufAnimation *animation;
+    GdkPixbufAnimationIter *iter;
+    GIOChannel       *io_channel;
     GdkPixbuf        *src_pixbuf;
     GdkPixbuf        *dst_pixbuf; /* The pixbuf which ends up on screen */
     RsttoNavigator   *navigator;
@@ -57,9 +61,6 @@ rstto_picture_viewer_paint(GtkWidget *widget);
 static gboolean
 rstto_picture_viewer_refresh(RsttoPictureViewer *viewer);
 
-static void
-rstto_picture_viewer_update(RsttoPictureViewer *viewer);
-
 static gboolean
 rstto_picture_viewer_set_scroll_adjustments(RsttoPictureViewer *, GtkAdjustment *, GtkAdjustment *);
 
@@ -72,6 +73,16 @@ cb_rstto_picture_viewer_scroll_event (RsttoPictureViewer *, GdkEventScroll *);
 static gboolean
 cb_rstto_picture_viewer_update_image(RsttoPictureViewer *viewer);
 
+static gboolean
+cb_rstto_picture_viewer_read_file(GIOChannel *io_channel, GIOCondition cond, RsttoPictureViewer *viewer);
+
+static void
+cb_rstto_picture_viewer_area_prepared(GdkPixbufLoader *loader, RsttoPictureViewer *viewer);
+static void
+cb_rstto_picture_viewer_area_updated(GdkPixbufLoader *loader, gint x, gint y, gint width, gint height, RsttoPictureViewer *viewer);
+static void
+cb_rstto_picture_viewer_closed(GdkPixbufLoader *loader, RsttoPictureViewer *viewer);
+
 
 static GtkWidgetClass *parent_class = NULL;
 
@@ -236,10 +247,6 @@ rstto_picture_viewer_paint(GtkWidget *widget)
     /* required for transparent pixbufs... add double buffering to fix flickering*/
     if(GTK_WIDGET_REALIZED(widget))
     {          
-        GdkCursor *cursor = gdk_cursor_new(GDK_WATCH);
-        gdk_window_set_cursor(widget->window, cursor);
-        gdk_cursor_unref(cursor);
-
         GdkPixmap *buffer = gdk_pixmap_new(NULL, widget->allocation.width, widget->allocation.height, gdk_drawable_get_depth(widget->window));
         GdkGC *gc = gdk_gc_new(GDK_DRAWABLE(buffer));
 
@@ -322,9 +329,7 @@ rstto_picture_viewer_paint(GtkWidget *widget)
                         widget->allocation.width,
                         widget->allocation.height);
         g_object_unref(buffer);
-
-        gdk_window_set_cursor(widget->window, NULL);
-    }
+   }
 }
 
 static void
@@ -380,7 +385,7 @@ cb_rstto_picture_viewer_value_changed(GtkAdjustment *adjustment, RsttoPictureVie
     GdkPixbuf *tmp_pixbuf = NULL;
     if (viewer->vadjustment && viewer->hadjustment)
     {
-        if (scale < 1.0)
+        if ((width > widget->allocation.width) || (height > widget->allocation.height))
         {
             tmp_pixbuf = gdk_pixbuf_new_subpixbuf(viewer->priv->src_pixbuf,
                                               viewer->hadjustment->value / scale >= 0?
@@ -452,8 +457,8 @@ rstto_picture_viewer_set_scale(RsttoPictureViewer *viewer, gdouble scale)
     RsttoNavigatorEntry *entry = rstto_navigator_get_file(viewer->priv->navigator);
     if (entry)
     {
-        rstto_navigator_entry_set_scale(entry, scale);
         rstto_navigator_entry_set_fit_to_screen (entry, FALSE);
+        rstto_navigator_entry_set_scale(entry, scale);
 
         if(rstto_picture_viewer_refresh(viewer))
         {
@@ -507,6 +512,13 @@ rstto_picture_viewer_refresh(RsttoPictureViewer *viewer)
 
     gboolean vadjustment_changed = FALSE;
     gboolean hadjustment_changed = FALSE;
+
+    if (viewer->priv->src_pixbuf != NULL && entry == NULL)
+    {
+        gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+        viewer->priv->src_pixbuf = NULL;
+    }
+
     if(viewer->priv->src_pixbuf)
     {
         gdouble width = (gdouble)gdk_pixbuf_get_width(viewer->priv->src_pixbuf);
@@ -551,10 +563,6 @@ rstto_picture_viewer_refresh(RsttoPictureViewer *viewer)
         }
         if(GTK_WIDGET_REALIZED(widget))
         {
-            GdkCursor *cursor = gdk_cursor_new(GDK_WATCH);
-            gdk_window_set_cursor(widget->window, cursor);
-            gdk_cursor_unref(cursor);
-
             gdouble width = (gdouble)gdk_pixbuf_get_width(viewer->priv->src_pixbuf);
             gdouble height = (gdouble)gdk_pixbuf_get_height(viewer->priv->src_pixbuf);
             
@@ -634,8 +642,6 @@ rstto_picture_viewer_refresh(RsttoPictureViewer *viewer)
                 gtk_adjustment_value_changed(viewer->hadjustment);
             if (vadjustment_changed == TRUE)
                 gtk_adjustment_value_changed(viewer->vadjustment);
-                
-            gdk_window_set_cursor(widget->window, NULL);
         }
     }
     else
@@ -652,15 +658,51 @@ rstto_picture_viewer_refresh(RsttoPictureViewer *viewer)
 static gboolean
 cb_rstto_picture_viewer_update_image(RsttoPictureViewer *viewer)
 {
-    gdouble sec;
-
-    sec = g_timer_elapsed(viewer->priv->timer, NULL);
-    if (sec >= 0.250)
+    if (viewer->priv->iter)
     {
-        g_timer_stop(viewer->priv->timer);
-        viewer->priv->timeout_id = 0;
-        rstto_picture_viewer_update(viewer);
-        return FALSE;
+        if(gdk_pixbuf_animation_iter_advance(viewer->priv->iter, NULL))
+        {
+            /* Cleanup old image */
+            if (viewer->priv->src_pixbuf)
+            {
+                gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+                viewer->priv->src_pixbuf = NULL;
+            }
+            RsttoNavigatorEntry *entry = rstto_navigator_get_file(viewer->priv->navigator);
+
+            if (entry)
+            {
+                GdkPixbuf *src_pixbuf = gdk_pixbuf_animation_iter_get_pixbuf(viewer->priv->iter);
+                if (src_pixbuf)
+                {
+                    viewer->priv->src_pixbuf = gdk_pixbuf_rotate_simple(src_pixbuf, rstto_navigator_entry_get_rotation(entry));
+                    if (rstto_navigator_entry_get_flip(entry, FALSE))
+                    {
+                        src_pixbuf = viewer->priv->src_pixbuf;
+                        viewer->priv->src_pixbuf = gdk_pixbuf_flip(src_pixbuf, FALSE);
+                        gdk_pixbuf_unref(src_pixbuf);
+                    }
+
+                    if (rstto_navigator_entry_get_flip(entry, TRUE))
+                    {
+                        src_pixbuf = viewer->priv->src_pixbuf;
+                        viewer->priv->src_pixbuf = gdk_pixbuf_flip(src_pixbuf, TRUE);
+                        gdk_pixbuf_unref(src_pixbuf);
+                    }
+                }
+            }
+
+            rstto_picture_viewer_refresh(viewer);
+            rstto_picture_viewer_paint(GTK_WIDGET(viewer));
+
+            gint time = gdk_pixbuf_animation_iter_get_delay_time(viewer->priv->iter);
+            if (time != -1)
+            {
+                viewer->priv->timeout_id = g_timeout_add(time, (GSourceFunc)cb_rstto_picture_viewer_update_image, viewer);
+            }
+
+            return FALSE;
+        }
     }
     return TRUE;
 }
@@ -670,40 +712,45 @@ cb_rstto_picture_viewer_nav_iter_changed(RsttoNavigator *nav, gint nr, RsttoNavi
 {
     GtkWidget *widget = GTK_WIDGET(viewer);
     if(entry)
-    {
+    { 
         if (GTK_WIDGET_REALIZED(widget))
         {
             GdkCursor *cursor = gdk_cursor_new(GDK_WATCH);
             gdk_window_set_cursor(widget->window, cursor);
             gdk_cursor_unref(cursor);
         }
-        if (viewer->priv->timeout == TRUE)
+        if (viewer->priv->loader)
         {
-            if (viewer->priv->timeout_id == 0)
+            g_signal_handlers_disconnect_by_func(viewer->priv->loader , cb_rstto_picture_viewer_area_prepared, viewer);
+            g_signal_handlers_disconnect_by_func(viewer->priv->loader , cb_rstto_picture_viewer_area_updated, viewer);
+            gdk_pixbuf_loader_close(viewer->priv->loader, NULL);
+            g_object_unref(viewer->priv->loader);
+            if (viewer->priv->iter)
             {
-                if (viewer->priv->dst_pixbuf)
-                {
-                    GdkPixbuf *pixbuf = gdk_pixbuf_composite_color_simple (viewer->priv->dst_pixbuf,
-                                                                    gdk_pixbuf_get_width(viewer->priv->dst_pixbuf),
-                                                                    gdk_pixbuf_get_height(viewer->priv->dst_pixbuf),
-                                                                    GDK_INTERP_BILINEAR,
-                                                                    100,
-                                                                    100,
-                                                                    0x000000,
-                                                                    0x000000);
-                    gdk_pixbuf_unref(viewer->priv->dst_pixbuf);
-                    viewer->priv->dst_pixbuf = pixbuf;
-                                                                    
-                    rstto_picture_viewer_paint(GTK_WIDGET(viewer));
-                }
-                viewer->priv->timeout_id = g_timeout_add(100, (GSourceFunc)cb_rstto_picture_viewer_update_image, viewer);
+                viewer->priv->iter = NULL;
+            }
+            if(viewer->priv->animation)
+            {
+                g_object_unref(viewer->priv->animation);
+                viewer->priv->animation = NULL;
+            }
+            if(viewer->priv->src_pixbuf)
+            {
+                gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+                viewer->priv->src_pixbuf = NULL;
             }
-            g_timer_start(viewer->priv->timer);
-        }
-        else
-        {
-            rstto_picture_viewer_update(viewer);
         }
+        viewer->priv->loader = gdk_pixbuf_loader_new();
+
+        g_signal_connect(viewer->priv->loader, "area-prepared", G_CALLBACK(cb_rstto_picture_viewer_area_prepared), viewer);
+        g_signal_connect(viewer->priv->loader, "area-updated", G_CALLBACK(cb_rstto_picture_viewer_area_updated), viewer);
+        g_signal_connect(viewer->priv->loader, "closed", G_CALLBACK(cb_rstto_picture_viewer_closed), viewer);
+        ThunarVfsInfo *info = rstto_navigator_entry_get_info(entry);
+        gchar *path = thunar_vfs_path_dup_string(info->path);
+
+        viewer->priv->io_channel = g_io_channel_new_file(path, "r", NULL);
+        g_io_channel_set_encoding(viewer->priv->io_channel, NULL, NULL);
+        g_io_add_watch(viewer->priv->io_channel, G_IO_IN | G_IO_PRI, (GIOFunc)cb_rstto_picture_viewer_read_file, viewer);
     }
     else
     {
@@ -724,40 +771,6 @@ cb_rstto_picture_viewer_nav_iter_changed(RsttoNavigator *nav, gint nr, RsttoNavi
     }
 }
 
-static void
-rstto_picture_viewer_update(RsttoPictureViewer *viewer)
-{
-    RsttoNavigatorEntry *entry = rstto_navigator_get_file(viewer->priv->navigator);
-    GtkWidget *widget = GTK_WIDGET(viewer);
-    if(GTK_WIDGET_REALIZED(widget))
-    {
-        GdkCursor *cursor = gdk_cursor_new(GDK_WATCH);
-        gdk_window_set_cursor(widget->window, cursor);
-        gdk_cursor_unref(cursor);
-    }
-
-    if(viewer->priv->src_pixbuf)
-    {
-        g_object_unref(viewer->priv->src_pixbuf);
-        viewer->priv->src_pixbuf = NULL;
-    }
-
-    if(entry)
-    {
-        viewer->priv->src_pixbuf = rstto_navigator_entry_get_pixbuf(entry);
-        if (viewer->priv->src_pixbuf)
-            g_object_ref(viewer->priv->src_pixbuf);
-    }
-
-    if(GTK_WIDGET_REALIZED(widget))
-    {
-        rstto_picture_viewer_refresh(viewer);
-        rstto_picture_viewer_paint(GTK_WIDGET(viewer));
-
-        gdk_window_set_cursor(widget->window, NULL);
-    }
-}
-
 static void
 cb_rstto_picture_viewer_scroll_event (RsttoPictureViewer *viewer, GdkEventScroll *event)
 {
@@ -795,3 +808,137 @@ rstto_picture_viewer_set_timeout(RsttoPictureViewer *viewer, gboolean timeout)
 {
     viewer->priv->timeout = timeout;
 }
+
+static gboolean
+cb_rstto_picture_viewer_read_file(GIOChannel *io_channel, GIOCondition cond, RsttoPictureViewer *viewer)
+{
+    gchar buffer[1024];
+    gsize bytes_read = 0;
+    GError *error = NULL;
+
+
+    GIOStatus status;
+    if (viewer->priv->io_channel == io_channel)
+    {
+        status = g_io_channel_read_chars(io_channel, buffer, 1024, &bytes_read,  &error);
+
+        switch (status)
+        {
+            case G_IO_STATUS_NORMAL:
+                if(gdk_pixbuf_loader_write(viewer->priv->loader, (const guchar *)buffer, bytes_read, NULL) == FALSE)
+                {
+                    gdk_pixbuf_loader_close(viewer->priv->loader, NULL);
+                    viewer->priv->io_channel = NULL;
+                    return FALSE;
+                }
+                return TRUE;
+                break;
+            case G_IO_STATUS_EOF:
+                gdk_pixbuf_loader_write(viewer->priv->loader, (const guchar *)buffer, bytes_read, NULL);
+                gdk_pixbuf_loader_close(viewer->priv->loader, NULL);
+                viewer->priv->io_channel = NULL;
+                return FALSE;
+                break;
+            case G_IO_STATUS_ERROR:
+                gdk_pixbuf_loader_close(viewer->priv->loader, NULL);
+                viewer->priv->io_channel = NULL;
+                return FALSE;
+                break;
+            case G_IO_STATUS_AGAIN:
+                return TRUE;
+                break;
+        }
+    }
+    g_io_channel_shutdown(io_channel, FALSE, NULL);
+    return FALSE;
+}
+
+static void
+cb_rstto_picture_viewer_area_prepared(GdkPixbufLoader *loader, RsttoPictureViewer *viewer)
+{
+    viewer->priv->animation = gdk_pixbuf_loader_get_animation(loader);
+    viewer->priv->iter = gdk_pixbuf_animation_get_iter(viewer->priv->animation, NULL);
+    if (viewer->priv->src_pixbuf)
+    {
+        gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+        viewer->priv->src_pixbuf = NULL;
+    }
+
+    gint time = gdk_pixbuf_animation_iter_get_delay_time(viewer->priv->iter);
+
+    rstto_picture_viewer_refresh(viewer);
+    rstto_picture_viewer_paint(GTK_WIDGET(viewer));
+
+    if (time != -1)
+    {
+        /* update frame */
+        viewer->priv->timeout_id = g_timeout_add(time, (GSourceFunc)cb_rstto_picture_viewer_update_image, viewer);
+    }   
+    else
+    {
+        viewer->priv->iter = NULL;
+    }
+}
+
+static void
+cb_rstto_picture_viewer_area_updated(GdkPixbufLoader *loader, gint x, gint y, gint width, gint height, RsttoPictureViewer *viewer)
+{
+    if (viewer->priv->iter)
+    {
+        /* Current Frame being updated?! */
+        if (gdk_pixbuf_animation_iter_on_currently_loading_frame(viewer->priv->iter) == TRUE)
+        {
+            /* Is it inside the viewport? */
+            RsttoNavigatorEntry *entry = rstto_navigator_get_file(viewer->priv->navigator);
+            gdouble scale = rstto_navigator_entry_get_scale(entry);
+            if (((viewer->vadjustment->value > (y * scale)) && (viewer->vadjustment->value < ((y+height) * scale))) &&
+                ((viewer->hadjustment->value > (x * scale)) && (viewer->hadjustment->value < ((x+width) * scale))))
+            {
+            
+            /* Update */
+            /*
+            gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+            viewer->priv->src_pixbuf = gdk_pixbuf_animation_iter_get_pixbuf(viewer->priv->iter);
+            gdk_pixbuf_ref(viewer->priv->src_pixbuf);
+            rstto_picture_viewer_refresh(viewer);
+            rstto_picture_viewer_paint(viewer);
+            */
+            }
+        }
+    }
+}
+
+static void
+cb_rstto_picture_viewer_closed(GdkPixbufLoader *loader, RsttoPictureViewer *viewer)
+{
+    GtkWidget *widget = GTK_WIDGET(viewer);
+
+    if (viewer->priv->src_pixbuf)
+    {
+        gdk_pixbuf_unref(viewer->priv->src_pixbuf);
+        viewer->priv->src_pixbuf = NULL;
+    }
+    if (viewer->priv->iter)
+    {
+        viewer->priv->src_pixbuf = gdk_pixbuf_animation_iter_get_pixbuf(viewer->priv->iter);
+        viewer->priv->src_pixbuf = gdk_pixbuf_copy(viewer->priv->src_pixbuf);
+    }
+    else
+    {
+        if (viewer->priv->loader)
+        {
+            viewer->priv->src_pixbuf = gdk_pixbuf_loader_get_pixbuf(viewer->priv->loader);
+            if (viewer->priv->src_pixbuf)
+            {
+                gdk_pixbuf_ref(viewer->priv->src_pixbuf);
+            }
+        }
+    }
+
+    rstto_picture_viewer_refresh(viewer);
+    rstto_picture_viewer_paint(GTK_WIDGET(viewer));
+    if (GTK_WIDGET_REALIZED(widget))
+    {
+        gdk_window_set_cursor(widget->window, NULL);
+    }
+}
